#include "set2d.h"

/*****************************************************************************
*                         Functions of output                                *
*****************************************************************************/
void out_bunch(void)
{
#ifdef energy_total
  out_energy_total();
#endif
#ifdef thermalflux
  out_thermalflux();
#endif
#ifdef field_RMS
  if(lap % 20 == 0)
    out_field_RMS();
#endif
#ifdef field_magnetic_hourly
  out_field_magnetic_hourly();
#endif
#ifdef field_absorption
  out_field_absorption();
#endif

  if((lap % STEP == 0 && lap >= (int)BEGIN) || lap == 0)
    {
#ifdef particle_phase
      out_particle_phase();
#endif
#ifdef particle_density_pm3d
      out_particle_density_pm3d();
#endif
#ifdef field_electric
      out_field_electric();
#endif
#ifdef field_magnetic
      out_field_magnetic();
#endif
#ifdef energy_spectrum
      out_energy_spectrum();
#endif
#ifdef project_density
      out_project_density();
#endif
#ifdef project_laser
      out_project_laser();
#endif
#ifdef particle_distribution
      out_particle_distribution();
#endif
#ifdef energy_gain
      out_energy_gain();
#endif
    }
}

#ifdef energy_spectrum
#define egd 20000
#define igd 1000
#define dz  (dx)

void out_energy_spectrum(void)
{
  double abv[2][nmx];
  double cte[egd+1],fde[egd+1];
  double cti[igd+1],fdi[igd+1];
  double NC  = 4792038.417*(dx*dy*dz);
  double eVi = 469.3526407;
  double eVe = 0.5109990645;
  FILE *esfp,*isfp;

  sprintf(fname,"spe%04d.dat",(int)(lap*dt));
  esfp = fopen(fname,"w");
  sprintf(fname,"spi%04d.dat",(int)(lap*dt));
  isfp = fopen(fname,"w");
  
  for(j=0;j<=egd;j++)
    cte[j] = 0.0;
  for(j=0;j<=igd;j++)
    cti[j] = 0.0;
  
  for(i=0;i<2;i++)
    for(j=0;j<nmx;j++)
      abv[i][j] = sqrt(vav(vx[i][j],vy[i][j],vz[i][j]));

  for(j=0;j<nmx;j++)
    {
      cte[(int)(abv[0][j]*egd)]++;
      cti[(int)(abv[1][j]*igd)]++;
    }

  for(j=0;j<egd;j++)
    fde[j] = 1.0/sqrt((1.0-(((double)(j)/egd)*((double)(j)/egd))));
  for(j=0;j<igd;j++)
    fdi[j] = ((double)(j)/igd)*((double)(j)/igd);
  
  for(i=0;i<egd;i++)
    fprintf(esfp,"%e\t%e\t%e\n",eVe*(fde[i]-1.0),cte[i]/NC,cte[i]*norm);
  for(i=0;i<igd;i++)
    fprintf(isfp,"%e\t%e\t%e\n",eVi*fdi[i],cti[i]/NC,cti[i]*norm);
  
  fclose(isfp); 
  fclose(esfp);
}
#endif

#ifdef energy_total
void out_energy_total(void)
{
  double tk[2],sum,Loz;
  double ebd[2],bzs,exs,eys;
  FILE *tefp;

  if(lap == 0)
    tefp = fopen("energy.dat","w");
  else 
    tefp = fopen("energy.dat","a");
  
  tk[0] = tk[1] = bzs = exs = eys = ebd[0] = ebd[1] = 0.0; 

  for(i=0;i<2;i++)
    for(j=0;j<nmx;j++)
      {
	Loz = sqrt(1.0/(1.0 - vav(vx[i][j],vy[i][j],vz[i][j])));
	tk[i] += m[i]*(Loz - 1.0)*norm/tg;
      }
  
  for(i=1;i<=imx;i++)
    for(j=0;j<=jmx;j++)
      exs += (0.5*Ex[i][j]*Ex[i][j])/tg;
  
  for(i=0;i<=imx;i++)
    for(j=1;j<=jmx;j++)
      eys += (0.5*Ey[i][j]*Ey[i][j])/tg;
  
  for(i=1;i<=imx;i++)
    for(j=1;j<=jmx;j++)
      bzs += (0.5*Bz[i][j]*Bz[i][j])/tg;
  
  for(i=1;i<=imx;i++)
    ebd[0] += 0.125*(Ex[i][0]+Ex[i][jmx])*(Ex[i][0]+Ex[i][jmx])/tg;
  
  for(j=1;j<=jmx;j++)
    ebd[1] += 0.125*(Ey[0][j]+Ex[imx][j])*(Ex[0][j]+Ex[imx][j])/tg;
  
  sum = tk[0] + tk[1] + bzs + exs + eys - ebd[0] - ebd[1];

  fprintf(tefp,"%f\t%e\t%e\t%e\t%e\t%e\t%e\n",(double)(lap*dt),
	  tk[0],tk[1],exs-ebd[0],eys-ebd[1],bzs,sum);
  fclose(tefp);
}
#endif

#ifdef particle_density_pm3d
void out_particle_density_pm3d(void)
{
  double dn[imx+1][jmx+1];
  FILE  *dnfp;

  /********* Particle of electron **********/
  sprintf(fname,"dne%04d.dat",(int)(lap*dt));
  dnfp = fopen(fname,"w");

  for(i=0;i<=imx;i++)
    for(j=0;j<=jmx;j++)
      dn[i][j] = 0.0;
  
  for(j=0;j<nmx;j++)
    dn[(int)((x[0][j]+hdx)/dx)][(int)((y[0][j]+hdy)/dy)]++;
  
  for(i=1;i<=imx;i++){
    for(j=1;j<=jmx;j++)
      fprintf(dnfp,"%e\t%e\t%e\n",
	      ((i-1)*dx+hdx)/twp,((j-1)*dy+hdy)/twp,dn[i][j]*norm);
    fprintf(dnfp,"\n");
  }
  fclose(dnfp);

  /********* Particle of ion **********/
  sprintf(fname,"dni%04d.dat",(int)(lap*dt));
  dnfp = fopen(fname,"w");

  for(i=0;i<=imx;i++)
    for(j=0;j<=jmx;j++)
      dn[i][j] = 0.0;
  
  for(j=0;j<nmx;j++)
    dn[(int)((x[1][j]+hdx)/dx)][(int)((y[1][j]+hdy)/dy)]++;
  
  for(i=1;i<=imx;i++){
    for(j=1;j<=jmx;j++)
      fprintf(dnfp,"%e\t%e\t%e\n",
	      ((i-1)*dx+hdx)/twp,((j-1)*dy+hdy)/twp,dn[i][j]*norm);
    fprintf(dnfp,"\n");
  }
  fclose(dnfp);
}
#endif

#ifdef particle_phase
void out_particle_phase(void)
{
  int lmx = 65535;
  FILE *phfp;
  
  if(nmx < lmx) lmx = nmx; 
  /********** particle of electron **********/
  sprintf(fname,"phe%04d.dat",(int)(lap*dt));
  phfp = fopen(fname,"w");
  
  for(j=0;j<lmx;j++)
    fprintf(phfp,"%e\t%e\t%e\t%e\n",
	    x[0][j]/twp, y[0][j]/twp,vx[0][j],vy[0][j]);
  fclose(phfp);

  /********** particle of ion **********/
  sprintf(fname,"phi%04d.dat",(int)(lap*dt));
  phfp = fopen(fname,"w");
  
  for(j=0;j<lmx;j++)
    fprintf(phfp,"%e\t%e\t%e\t%e\n",
	    x[1][j]/twp, y[1][j]/twp,vx[1][j],vy[1][j]);
  fclose(phfp);
}
#endif

#ifdef thermalflux
void out_thermalflux(void)
{
  double tf1,tf2,tf3;
  double Loz,Avx,tal;
  double stv;
  FILE *thfp;

  if(lap == 0)
    thfp = fopen("thermal.dat","w");
  else 
    thfp = fopen("thermal.dat","a");

  Loz = Avx = tf1 = tf2 = tf3 = stv = 0.0;
  k = 0;
  
  for(j=0;j<nmx;j++){
    if(x[0][j] >= dbp && x[0][j] <= dbp + tlp)
      {
	Loz = 1.0/sqrt(1.0 - vav(vx[0][j],vy[0][j],vz[0][j]));
	Avx += vx[0][j]*Loz;
	stv += vx[0][j];
	k++;
      }
  }
  
  Avx = Avx/(double)(k);
  stv = stv/(double)(k);
  
  for(j=0;j<nmx;j++){
    if(x[0][j] >= dbp && x[0][j] <= dbp + tlp)
      {
	Loz = 1.0/sqrt(1.0 - vav(vx[0][j],vy[0][j],vz[0][j]));
	
	tal = vav((vx[0][j]*Loz -Avx),vy[0][j]*Loz,vz[0][j]*Loz);
	tf1 += (vx[0][j]*Loz -Avx)*tal;

	tal = vav(vx[0][j]*Loz,vy[0][j]*Loz,vz[0][j]*Loz);
	tf2 += vx[0][j]*Loz*tal;
	
	tf3 += 0.5*(vx[0][j]-stv)*vav(vx[0][j]-stv,vy[0][j],vz[0][j]);
	/*tf3 += (vx[0][j]*Loz)*(vx[0][j]*Loz)*(vx[0][j]*Loz);*/
      }
  }
  
  fprintf(thfp,"%e\t%e\t%e\t%e\n",(double)(lap*dt),tf1,tf2,tf3);
  fclose(thfp);
}
#endif

#ifdef field_electric
void out_field_electric(void)
{
  FILE *effp;
  
  /********* Electric field (x) **********/
  sprintf(fname,"efx%04d.dat",(int)(lap*dt));
  effp = fopen(fname,"w");
  
  for(i=1;i<=imx;i++){
    for(j=0;j<=jmx;j++)
      fprintf(effp,"%e\t%e\t%e\n",
	      ((i-1)*dx)/twp,((j-1)*dy+hdy)/twp,Ex[i][j]);
    fprintf(effp,"\n");
  }
  fclose(effp);
  
  /********* Electric field (y) **********/
  sprintf(fname,"efy%04d.dat",(int)(lap*dt));
  effp = fopen(fname,"w");
  
  for(i=0;i<=imx;i++){
    for(j=1;j<=jmx;j++)
      fprintf(effp,"%e\t%e\t%e\n",
	      ((i-1)*dx+hdx)/twp,((j-1)*dy)/twp,Ey[i][j]);
    fprintf(effp,"\n");
  }
  fclose(effp);
}
#endif

#ifdef field_magnetic
void out_field_magnetic(void)
{
  FILE *bffp;
  
  /********* Magnetic field (z) **********/
  sprintf(fname,"bfz%04d.dat",(int)(lap*dt));
  bffp = fopen(fname,"w");
  
  for(i=1;i<=imx;i++){
    for(j=1;j<=jmx;j++)
      fprintf(bffp,"%e\t%e\t%e\n",
	      ((i-1)*dx+hdx)/twp,((j-1)*dy+hdy)/twp,Bz[i][j]);
    fprintf(bffp,"\n");
  }
  fclose(bffp);
}
#endif

#ifdef project_laser
void out_project_laser(void)
{
  FILE *plfp;
  
  sprintf(fname,"prl%04d.dat",(int)(lap*dt));
  plfp = fopen(fname,"w");

  for(i=0;i<=imx;i++)
    fprintf(plfp,"%e %e %e %e\n",i*dx/twp,
	    Ey[i][(int)(jmx/2)],Ex[i][(int)(jmx/2)],Bz[i][(int)(jmx/2)]);

  fclose(plfp);
}
#endif

#ifdef project_density
void out_project_density(void)
{
  double pdn[imx+2];
  FILE *pdfp;
  
  sprintf(fname,"prd%04d.dat",(int)(lap*dt));
  pdfp = fopen(fname,"w");
  
  for(i=0;i<=imx;i++)
    pdn[i] = 0.0;

  for(j=0;j<nmx;j++)
    pdn[(int)((x[0][j]+hdx)/dx)]++;
  
  for(i=0;i<=imx;i++)
    fprintf(pdfp,"%e %e\n",i*dx/twp,pdn[i]/(double)(jmx)*norm);

  fclose(pdfp);
}
#endif

#ifdef field_RMS
void out_field_RMS(void)
{
  double Ism,Iyz;
  FILE *rmfp;
  
  if(lap == 0)
    rmfp = fopen("rms_Iyz.dat","w");
  else
    {
      rmfp = fopen("rms_Iyz.dat","a"); 
      fprintf(rmfp,"\n");
    }

  Ism = Iyz = 0.0;
  
  for(i=1;i<=imx;i++)
    for(j=1;j<=jmx;j++)
      Ism += (Ex[i][j]*Ex[i][j] + Ey[i][j]*Ey[i][j])/2.0;
  
  Ism = Ism / (double)(imx*jmx);
  
  for(i=1;i<=imx;i++)
    {
      Iyz = 0.0;
      
      for(j=1;j<=jmx;j++)
	Iyz += Ey[i][j]*Ey[i][j]/2.0;

      Iyz = Iyz / (double)(jmx);
      fprintf(rmfp,"%e %e %e\n",i*dx/twp,lap*dt,
	      sqrt((Ism-Iyz)*(Ism-Iyz)/(double)(jmx)));
    }

  fclose(rmfp);
}
#endif

#ifdef particle_distribution
#define smx 101
void out_particle_distribution(void)
{
  double cvx[2][smx];
  double cvy[2][smx];
  double cvz[2][smx];
  FILE *dsfp;
   
  sprintf(fname,"pde%04d.dat",(int)(lap*dt));
  dsfp = fopen(fname,"w");

  for(i=0;i<2;i++)
    for(j=0;j<smx;j++)
      cvx[i][j] = cvy[i][j] = cvz[i][j] = 0.0;
  
  for(i=0;i<2;i++)
    for(j=0;j<nmx;j++)
      {
	cvx[i][(int)((vx[i][j]+1)*50)]++;
	cvy[i][(int)((vy[i][j]+1)*50)]++;
	cvz[i][(int)((vz[i][j]+1)*50)]++;
      }
	
  for(i=0;i<2;i++)
    for(j=0;j<smx;j++)
      {
	cvx[i][j] /= (double)nmx;
	cvy[i][j] /= (double)nmx;
	cvz[i][j] /= (double)nmx;
      }

  for(i=0;i<smx;i++)
    fprintf(dsfp,"%e %e %e %e\n",
	    2*(double)i/(double)(smx-1)-1,cvx[0][i],cvy[0][i],cvz[0][i]);

  fclose(dsfp);
}
#endif

#ifdef energy_gain
void out_energy_gain(void)
{
  double eng[imx+1],cnt[imx+1];
  double mmx[imx+1],mmy[imx+1];
  double tfx[imx+1],sey[imx+1];
  double Loz,Avx,fg,tf1;
  FILE *egfp;
  
  sprintf(fname,"ege%04d.dat",(int)(lap*dt));
  egfp = fopen(fname,"w");
  
  for(i=0;i<=imx;i++)
    eng[i] = cnt[i] = mmx[i] = mmy[i] = tfx[i] = sey[i] = Avx = 0.0;
  
  for(j=0;j<nmx;j++)
    Avx += vx[0][j]/(double)(nmx);

  for(j=0;j<nmx;j++)
    {	
      Loz = sqrt(1.0/(1.0 - vav(vx[0][j],vy[0][j],vz[0][j])));
      tf1 = 0.5*vav(vx[0][j]*Loz-Avx,vy[0][j]*Loz,vz[0][j]*Loz);
      
      eng[(int)((x[0][j]+hdx)/dx)] += (Loz - 1.0)*m[0];
      mmx[(int)((x[0][j]+hdx)/dx)] += vx[0][j];
      mmy[(int)((x[0][j]+hdx)/dx)] += vy[0][j];
      tfx[(int)((x[0][j]+hdx)/dx)] += (vx[0][j]*Loz-Avx)*tf1;
      cnt[(int)((x[0][j]+hdx)/dx)] ++;
    }
  
  for(i=1;i<=imx;i++)
    {
      if((int)(cnt[i]) == 0){
	cnt[i] = 1.0;
	fg     = 0.0;
      }else fg = 1.0;
      
      eng[i] = eng[i]/cnt[i]*fg;
      mmx[i] = mmx[i]/cnt[i]*fg;
      mmy[i] = mmy[i]/cnt[i]*fg;
    }
  
  for(i=0;i<=imx;i++)
    for(j=1;j<=jmx;j++)
      sey[i] += 0.5*Ey[i][j]*Ey[i][j]/(double)(jmx);

  for(i=1;i<=imx;i++)
    fprintf(egfp,"%e %e %e %e %e %e %e\n",
	    ((i-1)*dx+hdx)/twp,mmx[i],mmy[i],eng[i],tfx[i],cnt[i],sey[i]);
  
  fclose(egfp);
}
#endif

#ifdef field_magnetic_hourly
#define btu 100.0

void out_field_magnetic_hourly(void)
{
  FILE *bhfp;
  
  if(lap == 0)
    {
      for(i=0;i<=imx;i++)
        for(j=0;j<=jmx;j++)
	  Bhz[i][j] = 0.0;
    }
  
  /********* Magnetic field (z) **********/
 if(lap % (int)(btu) == 0 && lap != 0)
   {
     sprintf(fname,"bhz%04d.dat",(int)(lap*dt));
     bhfp = fopen(fname,"w");
     
     for(i=(int)((dbp)/dx);i<=(int)((dbp+tlp)/dx);i++){
       for(j=0;j<=jmx;j++)
	 fprintf(bhfp,"%e %e %e\n",i*dx/twp,j*dy/twp,Bhz[i][j]/btu);
       fprintf(bhfp,"\n");
     }

     fclose(bhfp);
     
     for(i=0;i<=imx;i++)
        for(j=0;j<=jmx;j++)
          Bhz[i][j] = 0.0;
   }
 
 for(i=0;i<=imx;i++)
   for(j=0;j<=jmx;j++)
     Bhz[i][j] += Bz[i][j];
}
#endif

#ifdef field_absorption
void out_field_absorption(void)
{
  FILE *abfp;
  double pyx[7],pyy[7];
  double sup[7],sum[7];
  static double tpp[7],tpm[7];
  int    ix[7];
  
  for(i=0;i<7;i++)
    sup[i] = sum[i] = 0.0;
  
  for(i=0;i<7;i++)
    for(j=1;j<=jmx;j++)
      {
	ix[i] = (int)(((dbp - um) + (um * i))/dx);
	if(i == 0) ix[i] = 1;

	pyx[i] = Ey[ix[i]][j]*Bz[ix[i]][j];
	pyy[i] = Ex[ix[i]][j]*Bz[ix[i]][j];

	if(pyx[i] >= 0)
	  sup[i] += sqrt(pyx[i]*pyx[i] + pyy[i]*pyy[i]);
	else
	  sum[i] += sqrt(pyx[i]*pyx[i] + pyy[i]*pyy[i]);
    }

  for(i=0;i<7;i++)
    {
      tpp[i] += sup[i];
      tpm[i] += sum[i];
    }

  if(lap == 0)
    {
      for(i=0;i<7;i++)
	tpp[i] = tpm[i] = 0.0;

      abfp = fopen("absorbe.dat","w");
      fclose(abfp);
    }
  
  if(lap == tmx)
    {
      abfp = fopen("absorbe.dat","a");
      
      for(i=0;i<7;i++)
	fprintf(abfp,"%e %e %e %e\n",ix[i]*dx/twp,
		(1.0-(tpm[i]/tpp[i]))*100,tpp[i],tpm[i]);
      fclose(abfp);
    }
}
#endif

/*************************** Simulation of setup *****************************/
#ifdef system_setup
void out_system_setup(void)
{
  FILE *sfp;
  sprintf(fname,"setup.dat");
  sfp = fopen(fname,"w");
  
  fprintf(sfp,"Setup of PIC Simulation \n");
  fprintf(sfp,"dx    = %f\n",   dx);
  fprintf(sfp,"dy    = %f\n",   dy);
  fprintf(sfp,"dt    = %f\n",   dt);
  fprintf(sfp,"imx   = %d\n",  imx);
  fprintf(sfp,"jmx   = %d\n",  jmx);
  fprintf(sfp,"Lx    = %f\n",   Lx);
  fprintf(sfp,"Ly    = %f\n",   Ly);
  fprintf(sfp,"tg    = %d\n",   tg);
  fprintf(sfp,"dt_dx = %f\n",dt_dx(dt));
  fprintf(sfp,"TL    = %d\n",  tmx);

  fprintf(sfp,"nmx   = %d\n",  nmx);
  fprintf(sfp,"vtxe  = %f\n", vtxe);
  fprintf(sfp,"vtye  = %f\n", vtye);
  fprintf(sfp,"vtze  = %f\n", vtze);
  fprintf(sfp,"vdxe  = %f\n", vdxe);
  fprintf(sfp,"vdye  = %f\n", vdye);
  fprintf(sfp,"vdze  = %f\n", vdze);
  fprintf(sfp,"vtxi  = %f\n", vtxi);
  fprintf(sfp,"vtyi  = %f\n", vtyi);
  fprintf(sfp,"vtzi  = %f\n", vtzi);
  fprintf(sfp,"vdxi  = %f\n", vdxi);
  fprintf(sfp,"vdyi  = %f\n", vdyi);
  fprintf(sfp,"vdzi  = %f\n", vdzi);

  fprintf(sfp,"\nParamerter of laser \n");
  fprintf(sfp,"Eo    = %f\n",   Eo);
  fprintf(sfp,"tau   = %f\n",  tau);
  fprintf(sfp,"w     = %f\n",    w);
  fprintf(sfp,"int llt   = %d\n",(int)llt);
  fprintf(sfp,"double llt  = %f\n",   llt);

  fprintf(sfp,"\nParameter of Plasma wall\n");
  fprintf(sfp,"tlp   = %f\n",  tlp);
#ifdef gradient
  fprintf(sfp,"lp    = %f\n",   lp);
  fprintf(sfp,"lg    = %f\n",   lg);
#else
  fprintf(sfp,"lp    = %f\n",   lp);
#endif
  fprintf(sfp,"dbp   = %f\n",  dbp);
  fprintf(sfp,"no    = %f\n",   no);
  fprintf(sfp,"np    = %f\n",   np);
  fprintf(sfp,"norm  = %f\n", norm);
  
  fclose(sfp);
}
#endif

/*****************************************************************************
*                             Utility Functions                              *
*****************************************************************************/
double grad(double width,double adside)
{
  double rx,ry;
  
  for(;;)
    {
      rx = rnd(1);
      ry = rnd(1); 
      if(rx > ry) break;
    }
  return width*rx + adside;
}

double vav(double tx,double ty,double tz)
{
  return (tx*tx)+(ty*ty)+(tz*tz);          /* Vector absolute value         */
}

double rnd(double tx)
{
  return tx*rand()/(RAND_MAX+1.0);         /* Function of rnd               */
}

double dt_dx(double tx)
{
  return (tx - dx)/(tx + dx);
}
